/*
 * Copyright (C) 2012 Christian Franke <nobody@nowhere.ws>
 * GPLv3+
 */

#define _GNU_SOURCE 1

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <event2/event.h>
#include <event2/buffer.h>
#include <event2/bufferevent.h>
#include <event2/http.h>

struct dht11_client {
	struct event_base *base;
	char *error;
	float temperature;
	float humidity;
	int fd;
	struct bufferevent *event;
	struct event *timeout;
};

#define DATA_PREFIX "Data: "
#define ERROR_PREFIX "Error: "

static void do_abort(struct event_base *evb, const char *msg)
{
	fprintf(stderr, "%s", msg);
	if (event_base_loopbreak(evb)) {
		fprintf(stderr, "Could not break event loop. Exit hard!\n");
		_exit(1);
	}
}

static void dht11_client_read(struct bufferevent *event, void *arg)
{
	struct dht11_client *dc = arg;
	char *line = evbuffer_readln(bufferevent_get_input(event),
			NULL, EVBUFFER_EOL_CRLF);
	if (!line)
		return;

	if (!strncmp(line, DATA_PREFIX, strlen(DATA_PREFIX))) {
		const char *data = line + strlen(DATA_PREFIX);
		if (sscanf(data, "%f, %f", &dc->temperature, &dc->humidity) == 2) {
			struct timeval tv = {
				.tv_sec = 30
			};

			free(dc->error);
			dc->error = NULL;
			free(line);

			if (evtimer_pending(dc->timeout, NULL) &&
			    evtimer_del(dc->timeout)) {
				do_abort(dc->base, "Could not clear old timer\n");
			}

			if (evtimer_add(dc->timeout, &tv))
				do_abort(dc->base, "Could not add new timer\n");

			return;
		}
	}

	/* We did not get data :( */
	free(dc->error);
	if (!strncmp(line, ERROR_PREFIX, strlen(ERROR_PREFIX))) {
		dc->error = strdup(line + strlen(ERROR_PREFIX));
	} else {
		if (asprintf(&dc->error, "Protocol (%s)", line) < 0)
			dc->error = NULL;
	}
	
	if (!dc->error)
		do_abort(dc->base, "Memory allocation failure.\n");

	free(line);
}

static void dht11_client_event(struct bufferevent *event, short what, void *arg)
{
	struct dht11_client *dc = arg;
	do_abort(dc->base, "Error on tty.\n");
}

static void dht11_client_timeout(evutil_socket_t fd, short what, void *arg)
{
	struct dht11_client *dc = arg;

	if (dc->error)
		return;

	dc->error = strdup("Outdated");
	if (!dc->error)
		do_abort(dc->base, "Memory allocation failure.\n");
}

static struct dht11_client *dht11_client_new(struct event_base *evb, const char *tty)
{
	struct dht11_client *rv = calloc(sizeof(*rv), 1);
	
	if (!rv)
		return NULL;

	rv->base = evb;
	rv->fd = open(tty, O_RDWR | O_NOCTTY | O_NDELAY);
	if (rv->fd < 0)
		goto out;

	rv->error = strdup("No Data Yet");
	if (!rv->error)
		goto out;

	if (isatty(rv->fd)) {
		struct termios options;

		if (tcgetattr(rv->fd, &options) < 0)
			goto out;

		if (cfsetispeed(&options, B9600) ||
		    cfsetospeed(&options, B9600))
			goto out;
		
		options.c_cflag |= CLOCAL | CREAD;
		options.c_lflag &= ~(ICANON|ECHO|ECHOE|ECHOK|ECHONL|ISIG|IEXTEN);
		options.c_lflag &= ~(ECHOCTL|ECHOKE);
		options.c_oflag &= ~(OPOST);
		options.c_iflag &= ~(INLCR|IGNCR|ICRNL|IGNBRK);
		options.c_iflag &= ~IUCLC;
		options.c_iflag &= ~PARMRK;
		options.c_cflag &= ~CSIZE;
		options.c_cflag |= CS8;
		options.c_cflag &= ~(CSTOPB);
		options.c_iflag &= ~(INPCK|ISTRIP);
		options.c_cflag &= ~(PARENB|PARODD);
		options.c_iflag &= ~(IXON|IXOFF|IXANY);
		options.c_cflag &= ~(CRTSCTS);
		options.c_cc[VMIN]= 0;
		options.c_cc[VTIME]= 0;
	
		if (tcsetattr(rv->fd, TCSANOW, &options))
			goto out;
	}

	rv->event = bufferevent_socket_new(evb, rv->fd, BEV_OPT_CLOSE_ON_FREE);
	if (!rv->event)
		goto out;

	bufferevent_setcb(rv->event, dht11_client_read, NULL, dht11_client_event, rv);
	if (bufferevent_enable(rv->event, EV_READ))
		goto out;

	rv->timeout = evtimer_new(evb, dht11_client_timeout, rv);
	if (!rv->timeout)
		goto out;

	return rv;

out:
	if (rv->fd >= 0)
		close(rv->fd);
	if (rv->event)
		bufferevent_free(rv->event);
	free(rv);
	return NULL;
}

static void handle_request(struct evhttp_request *req, void *arg)
{
	struct dht11_client *dc = arg;

	struct evbuffer *buf = evbuffer_new();
	if (!buf) {
		evhttp_send_error(req, HTTP_INTERNAL, "Memory allocation failed.");
		return;
	}

	int rv = 0;
	if (dc->error) {
		rv = evbuffer_add_printf(buf, "error: %s\n", dc->error);
	} else {
		rv = evbuffer_add_printf(buf, "temperature: %f\nhumidity: %f\n",
					 dc->temperature, dc->humidity);
	}

	if (rv < 0 ||
	    evhttp_add_header(evhttp_request_get_output_headers(req), "Content-Type",
		    "text/plain")) {
		evbuffer_free(buf);
		evhttp_send_error(req, HTTP_INTERNAL, "Could not format data.");
		return;
	}

	evhttp_send_reply(req, HTTP_OK, "Success",  buf);
	evbuffer_free(buf);
}

int main(int argc, char **argv)
{
	if (argc != 4) {
		fprintf(stderr, "Usage: %s <tty> <address> <port>\n", argv[0]);
		exit(1);
	}

	struct event_base *evb = event_base_new();
	if (!evb) {
		perror("event_base_new");
		exit(2);
	}

	struct dht11_client *dht = dht11_client_new(evb, argv[1]);
	if (!dht) {
		perror("dht11_client_new");
		exit(3);
	}

	struct evhttp *evh = evhttp_new(evb);
	if (!evh) {
		perror("evhttp_new");
		exit(4);
	}

	if (evhttp_set_cb(evh, "/", handle_request, dht)) {
		perror("evhttp_set_cb");
		exit(5);
	}

	if (evhttp_bind_socket(evh, argv[2], atoi(argv[3]))) {
		perror("evhttp_bind_socket");
		exit(6);
	}

	return event_base_dispatch(evb);
}
